@@ -1,6 +1,6 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2003, 2004, 2005, 2007, 2008, 
-   2009 Free Software Foundation, Inc.
+   Copyright (C) 2003-2005, 2007-2012, 2014-2016 Free Software
+   Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,9 +13,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with GNU Mailutils; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-   MA 02110-1301 USA */
+   along with GNU Mailutils.  If not, see <http://www.gnu.org/licenses/>. */
 
 #include "imap4d.h"
 #include <gsasl.h>
@@ -29,52 +27,23 @@ static Gsasl_session *sess_ctx;
 
 static void auth_gsasl_capa_init (int disable);
 
-static int
-create_gsasl_stream (mu_stream_t *newstr, mu_stream_t transport, int flags)
-{
-  int rc;
-  
-  rc = mu_gsasl_stream_create (newstr, transport, sess_ctx, flags);
-  if (rc)
-    {
-      mu_diag_output (MU_DIAG_ERROR, _("cannot create SASL stream: %s"),
-	      mu_strerror (rc));
-      return RESP_NO;
-    }
-
-  if ((rc = mu_stream_open (*newstr)) != 0)
-    {
-      const char *p;
-      if (mu_stream_strerror (*newstr, &p))
-	p = mu_strerror (rc);
-      mu_diag_output (MU_DIAG_ERROR, _("cannot open SASL input stream: %s"), p);
-      return RESP_NO;
-    }
-
-  return RESP_OK;
-}
-
-int
-gsasl_replace_streams (void *self, void *data)
-{
-  mu_stream_t *s = data;
-
-  util_set_input (s[0]);
-  util_set_output (s[1]);
-  free (s);
-  util_event_remove (self);
-  free (self);
-  return 0;
-}
-
 static void
-finish_session (void)
+finish_session (void *ptr)
 {
-  gsasl_finish (sess_ctx);
+  gsasl_finish ((Gsasl_session *)ptr);
 }
 
 static int
-auth_gsasl (struct imap4d_command *command, char *auth_type, char **username)
+restore_and_return (struct imap4d_auth *ap, mu_stream_t *str, int resp)
+{
+  mu_stream_unref (str[0]);
+  mu_stream_unref (str[1]);
+  ap->response = resp;
+  return imap4d_auth_resp;
+}
+
+static enum imap4d_auth_result
+auth_gsasl (struct imap4d_auth *ap)
 {
   char *input_str = NULL;
   size_t input_size = 0;
@@ -82,91 +51,169 @@ auth_gsasl (struct imap4d_command *command, char *auth_type, char **username)
   char *output;
   int rc;
   
-  rc = gsasl_server_start (ctx, auth_type, &sess_ctx);
+  rc = gsasl_server_start (ctx, ap->auth_type, &sess_ctx);
   if (rc != GSASL_OK)
     {
       mu_diag_output (MU_DIAG_NOTICE, _("SASL gsasl_server_start: %s"),
-	      gsasl_strerror(rc));
-      return 0;
+ 	              gsasl_strerror (rc));
+      return imap4d_auth_fail;
     }
 
-  gsasl_callback_hook_set (ctx, username);
+  gsasl_callback_hook_set (ctx, &ap->username);
 
   output = NULL;
   while ((rc = gsasl_step64 (sess_ctx, input_str, &output))
 	   == GSASL_NEEDS_MORE)
     {
-      util_send ("+ %s\r\n", output);
-      imap4d_getline (&input_str, &input_size, &input_len);
+      io_sendf ("+ %s\n", output);
+      io_getline (&input_str, &input_size, &input_len);
     }
   
-  free (input_str);
   if (rc != GSASL_OK)
     {
       mu_diag_output (MU_DIAG_NOTICE, _("GSASL error: %s"),
 		      gsasl_strerror (rc));
+      free (input_str);
       free (output);
-      return RESP_NO;
+      ap->response = RESP_NO;
+      return imap4d_auth_resp;
     }
 
-  /* Some SASL mechanisms output data when GSASL_OK is returned */
+  /* Some SASL mechanisms output additional data when GSASL_OK is
+     returned, and clients must respond with an empty response. */
   if (output[0])
-    util_send ("+ %s\r\n", output);
-  
+    {
+      io_sendf ("+ %s\n", output);
+      io_getline (&input_str, &input_size, &input_len);
+      if (input_len != 0)
+	{
+	  mu_diag_output (MU_DIAG_NOTICE, _("non-empty client response"));
+          free (input_str);
+          free (output);
+	  ap->response = RESP_NO;
+	  return imap4d_auth_resp;
+	}
+    }
+
+  free (input_str);
   free (output);
 
-  if (*username == NULL)
+  if (ap->username == NULL)
     {
-      mu_diag_output (MU_DIAG_NOTICE, _("GSASL %s: cannot get username"), auth_type);
-      return RESP_NO;
+      mu_diag_output (MU_DIAG_NOTICE, _("GSASL %s: cannot get username"),
+		      ap->auth_type);
+      ap->response = RESP_NO;
+      return imap4d_auth_resp;
     }
 
+  auth_gsasl_capa_init (1);
   if (sess_ctx)
     {
-      mu_stream_t tmp, new_in, new_out;
-      mu_stream_t *s;
+      mu_stream_t stream[2], newstream[2];
 
-      util_get_input (&tmp);
-      if (create_gsasl_stream (&new_in, tmp, MU_STREAM_READ))
-	return RESP_NO;
-      util_get_output (&tmp);
-      if (create_gsasl_stream (&new_out, tmp, MU_STREAM_WRITE))
+      rc = mu_stream_ioctl (iostream, MU_IOCTL_SUBSTREAM, MU_IOCTL_OP_GET, 
+                            stream);
+      if (rc)
 	{
-	  mu_stream_destroy (&new_in, mu_stream_get_owner (new_in));
-	  return RESP_NO;
+	  mu_error (_("%s failed: %s"), "MU_IOCTL_GET_STREAM",
+		    mu_stream_strerror (iostream, rc));
+	  ap->response = RESP_NO;
+	  return imap4d_auth_resp;
+	}
+      rc = gsasl_encoder_stream (&newstream[0], stream[0], sess_ctx,
+				 MU_STREAM_READ);
+      if (rc)
+	{
+	  mu_error (_("%s failed: %s"), "gsasl_encoder_stream",
+		    mu_strerror (rc));
+	  return restore_and_return (ap, stream, RESP_NO);
 	}
 
-      s = calloc (2, sizeof (mu_stream_t));
-      s[0] = new_in;
-      s[1] = new_out;
-      util_register_event (STATE_NONAUTH, STATE_AUTH,
-			   gsasl_replace_streams, s);
-      util_atexit (finish_session);
+      rc = gsasl_decoder_stream (&newstream[1], stream[1], sess_ctx,
+				 MU_STREAM_WRITE);
+      if (rc)
+	{
+	  mu_error (_("%s failed: %s"), "gsasl_decoder_stream",
+		    mu_strerror (rc));
+	  mu_stream_destroy (&newstream[0]);
+	  return restore_and_return (ap, stream, RESP_NO);
+	}
+      
+      if (ap->username)
+	{
+	  if (imap4d_session_setup (ap->username))
+	    {
+	      mu_stream_destroy (&newstream[0]);
+	      mu_stream_destroy (&newstream[1]);
+	      return restore_and_return (ap, stream, RESP_NO);
+	    }
+	}
+
+      /* FIXME: This is not reflected in the transcript. */
+      io_stream_completion_response (stream[1], ap->command, RESP_OK,
+				     "%s authentication successful",
+				     ap->auth_type);
+      mu_stream_flush (stream[1]);
+
+      mu_stream_unref (stream[0]);
+      mu_stream_unref (stream[1]);
+      
+      rc = mu_stream_ioctl (iostream, MU_IOCTL_SUBSTREAM, 
+                            MU_IOCTL_OP_SET, newstream);
+      if (rc)
+	{
+	  mu_error (_("%s failed when it should not: %s"),
+		    "MU_IOCTL_SET_STREAM",
+		    mu_stream_strerror (iostream, rc));
+	  abort ();
+	}
+
+      mu_stream_unref (newstream[0]);
+      mu_stream_unref (newstream[1]);
+      
+      mu_onexit (finish_session, sess_ctx);
+      return imap4d_auth_ok;
     }
   
-  auth_gsasl_capa_init (1);
-  return RESP_OK;
+  ap->response = RESP_OK;
+  return imap4d_auth_resp;
 }
 
 static void
 auth_gsasl_capa_init (int disable)
 {
   int rc;
-  char *listmech, *name, *s;
-
+  char *listmech;
+  struct mu_wordsplit ws;
+  
   rc =  gsasl_server_mechlist (ctx, &listmech);
   if (rc != GSASL_OK)
     return;
 
-  for (name = strtok_r (listmech, " ", &s); name;
-       name = strtok_r (NULL, " ", &s))
+  ws.ws_delim = " ";
+  if (mu_wordsplit (listmech, &ws,
+		    MU_WRDSF_DELIM|MU_WRDSF_SQUEEZE_DELIMS|
+		    MU_WRDSF_NOVAR|MU_WRDSF_NOCMD))
     {
-      if (disable)
-	auth_remove (name);
-      else
-	auth_add (strdup (name), auth_gsasl);
+      mu_error (_("cannot split line `%s': %s"), listmech,
+		mu_wordsplit_strerror (&ws));
+    }
+  else
+    {
+      size_t i;
+
+      for (i = 0; i < ws.ws_wordc; i++)
+	{
+	  if (disable)
+	    auth_remove (ws.ws_wordv[i]);
+	  else
+	    {
+	      auth_add (ws.ws_wordv[i], auth_gsasl);
+	      ws.ws_wordv[i] = NULL;
+	    }
+	}
+      mu_wordsplit_free (&ws);
     }
-      
   free (listmech);
 }
 
@@ -179,7 +226,7 @@ retrieve_password (Gsasl *ctx, Gsasl_session *sctx)
   const char *authid = gsasl_property_get (sctx, GSASL_AUTHID);
   
   if (username && *username == 0)
-    *username = strdup (authid);
+    *username = mu_strdup (authid);
 
   if (mu_gsasl_module_data.cram_md5_pwd
       && access (mu_gsasl_module_data.cram_md5_pwd, R_OK) == 0)
@@ -226,7 +273,7 @@ cb_validate (Gsasl *ctx, Gsasl_session *sctx)
   if (!pass)
     return GSASL_NO_PASSWORD;
   
-  *username = strdup (authid);
+  *username = mu_strdup (authid);
   
   auth = mu_get_auth_by_name (*username);
 
@@ -283,7 +330,7 @@ callback (Gsasl *ctx, Gsasl_session *sctx, Gsasl_property prop)
 	  char **username = gsasl_callback_hook_get (ctx);
 	  mu_diag_output (MU_DIAG_INFO, _("anonymous user %s logged in"),
 			  gsasl_property_get (sctx, GSASL_ANONYMOUS_TOKEN));
-	  *username = strdup (mu_gsasl_module_data.anon_user);
+	  *username = mu_strdup (mu_gsasl_module_data.anon_user);
 	}
       else
 	{
@@ -295,7 +342,7 @@ callback (Gsasl *ctx, Gsasl_session *sctx, Gsasl_property prop)
     case GSASL_VALIDATE_GSSAPI:
       {
 	char **username = gsasl_callback_hook_get (ctx);
-	*username = strdup (gsasl_property_get(sctx, GSASL_AUTHZID));
+	*username = mu_strdup (gsasl_property_get(sctx, GSASL_AUTHZID));
 	break;
       }
       
