@@ -1,6 +1,5 @@
-/*
-   GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2009, 2010 Free Software Foundation, Inc.
+/* GNU Mailutils -- a suite of utilities for electronic mail
+   Copyright (C) 2009-2012, 2014-2016 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -13,10 +12,8 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General
-   Public License along with this library; if not, write to the
-   Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-   Boston, MA 02110-1301 USA
-*/
+   Public License along with this library.  If not, see 
+   <http://www.gnu.org/licenses/>. */
 
 #include <mailutils/cpp/sieve.h>
 
@@ -28,7 +25,7 @@ using namespace mailutils;
 
 SieveMachine :: SieveMachine ()
 {
-  int status = mu_sieve_machine_init (&mach, NULL);
+  int status = mu_sieve_machine_init (&mach);
   if (status)
     throw Exception ("SieveMachine::SieveMachine", status);
 }
@@ -91,27 +88,4 @@ SieveMachine :: message (const Message& msg)
     throw Exception ("SieveMachine::message", status);
 }
 
-void
-SieveMachine :: set_debug (mu_sieve_printf_t printer)
-{
-  mu_sieve_set_debug (mach, printer);
-}
-
-void
-SieveMachine :: set_error (mu_sieve_printf_t printer)
-{
-  mu_sieve_set_error (mach, printer);
-}
-
-void
-SieveMachine :: set_parse_error (mu_sieve_parse_error_t printer)
-{
-  mu_sieve_set_parse_error (mach, printer);
-}
-
-void
-SieveMachine :: set_logger (mu_sieve_action_log_t printer)
-{
-  mu_sieve_set_logger (mach, printer);
-}
 
