@@ -1,6 +1,6 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2000, 2001, 2002, 2005,
-   2007, 2008, 2009 Free Software Foundation, Inc.
+   Copyright (C) 1999-2002, 2005, 2007-2012, 2014-2016 Free Software
+   Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,136 +13,125 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with GNU Mailutils; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-   MA 02110-1301 USA */
+   along with GNU Mailutils.  If not, see <http://www.gnu.org/licenses/>. */
 
 #include "maidag.h"
 
-void
-make_tmp (const char *from, mu_mailbox_t *mbox)
+static mu_message_t
+make_tmp (const char *from)
 {
-  struct mail_tmp *mtmp;
-  char *buf = NULL;
-  size_t n = 0;
   int rc;
-
-  if (mail_tmp_begin (&mtmp, from))
-    exit (EX_TEMPFAIL);
-
-  while (getline (&buf, &n, stdin) > 0)
-    if ((rc = mail_tmp_add_line (mtmp, buf, strlen (buf))))
-      break;
-  free (buf);
-  if (rc == 0)
-    rc = mail_tmp_finish (mtmp, mbox);
-  mail_tmp_destroy (&mtmp);
-  if (rc)
-    exit (EX_TEMPFAIL);
-}
-
-int
-mda (mu_mailbox_t mbx, char *username)
-{
-  int status;
-  mu_message_t msg;
+  mu_stream_t in, out;
+  char *buf = NULL;
+  size_t size = 0, n;
+  mu_message_t mesg;
   
-  if ((status = mu_mailbox_get_message (mbx, 1, &msg)) != 0)
+  rc = mu_stdio_stream_create (&in, MU_STDIN_FD, MU_STREAM_READ);
+  if (rc)
     {
-      maidag_error (_("cannot get message from the temporary mailbox: %s"),
-		    mu_strerror (status));
-      return EX_TEMPFAIL;
+      mu_diag_funcall (MU_DIAG_ERROR, "mu_stdio_stream_create",
+		       "MU_STDIN_FD", rc);
+      exit (EX_TEMPFAIL);
+    } 
+
+  rc = mu_temp_file_stream_create (&out, NULL, 0);
+  if (rc)
+    {
+      maidag_error (_("unable to open temporary file: %s"), mu_strerror (rc));
+      exit (EX_TEMPFAIL);
     }
 
-  deliver (msg, username, NULL);
+  rc = mu_stream_getline (in, &buf, &size, &n);
+  if (rc)
+    {
+      maidag_error (_("read error: %s"), mu_strerror (rc));
+      mu_stream_destroy (&in);
+      mu_stream_destroy (&out);
+      exit (EX_TEMPFAIL);
+    }
+  if (n == 0)
+    {
+      maidag_error (_("unexpected EOF on input"));
+      mu_stream_destroy (&in);
+      mu_stream_destroy (&out);
+      exit (EX_TEMPFAIL);
+    } 
 
-  if (multiple_delivery)
-    exit_code = EX_OK;
+  if (n >= 5 && memcmp (buf, "From ", 5))
+    {
+      struct mu_auth_data *auth = NULL;
+      if (!from)
+	{
+	  auth = mu_get_auth_by_uid (getuid ());
+	  if (auth)
+	    from = auth->name;
+	}
+      if (from)
+	{
+	  time_t t;
+	  struct tm *tm;
+	  
+	  time (&t);
+	  tm = gmtime (&t);
+	  mu_stream_printf (out, "From %s ", from);
+	  mu_c_streamftime (out, "%c%n", tm, NULL);
+	}
+      else
+	{
+	  maidag_error (_("cannot determine sender address"));
+	  mu_stream_destroy (&in);
+	  mu_stream_destroy (&out);
+	  exit (EX_TEMPFAIL);
+	}
+      if (auth)
+	mu_auth_data_free (auth);
+    }
 
-  return exit_code;
+  mu_stream_write (out, buf, n, NULL);
+  free (buf);
+  
+  rc = mu_stream_copy (out, in, 0, NULL);
+  mu_stream_destroy (&in);
+  if (rc)
+    {
+      maidag_error (_("copy error: %s"), mu_strerror (rc));
+      mu_stream_destroy (&out);
+      exit (EX_TEMPFAIL);
+    }
+
+  rc = mu_stream_to_message (out, &mesg);
+  mu_stream_destroy (&out);
+  if (rc)
+    {
+      maidag_error (_("error creating temporary message: %s"),
+		    mu_strerror (rc));
+      exit (EX_TEMPFAIL);
+    }
+
+  return mesg;
 }
 
 int
-maidag_stdio_delivery (int argc, char **argv)
+maidag_stdio_delivery (maidag_delivery_fn delivery_fun, int argc, char **argv)
 {
-  mu_mailbox_t mbox;
-
-  make_tmp (sender_address, &mbox);
+  mu_message_t mesg = make_tmp (sender_address);
   
   if (multiple_delivery)
     multiple_delivery = argc > 1;
 
   for (; *argv; argv++)
-    mda (mbox, *argv);
+    {
+      delivery_fun (mesg, *argv, NULL);
+      if (multiple_delivery)
+	exit_code = EX_OK;
+    }
   return exit_code;
 }
 
-static int biff_fd = -1;
-static struct sockaddr_in biff_in;
-static const char *biff_user_name;
-
-static int
-notify_action (mu_observer_t obs, size_t type, void *data, void *action_data)
-{
-  if (type == MU_EVT_MESSAGE_APPEND && biff_user_name)
-    {
-      mu_message_qid_t qid = data;
-      mu_mailbox_t mbox = mu_observer_get_owner (obs);
-      mu_url_t url;
-      char *buf;
-      
-      mu_mailbox_get_url (mbox, &url);
-      asprintf (&buf, "%s@%s:%s", biff_user_name,
-		qid, mu_url_to_string (url));
-      if (buf)
-	{
-	  sendto (biff_fd, buf, strlen (buf), 0,
-		  (struct sockaddr *)&biff_in, sizeof biff_in);
-	  free (buf);
-	}
-    }
-  return 0;
-}
-
-static void
-attach_notify (mu_mailbox_t mbox)
-{
-  struct servent *sp;
-  mu_observer_t observer;
-  mu_observable_t observable;
-
-  if (biff_fd == -1)
-    {
-      if ((sp = getservbyname ("biff", "udp")) == NULL)
-	{
-	  biff_fd = -2;
-	  return;
-	}
-      biff_in.sin_family = AF_INET;
-      biff_in.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
-      biff_in.sin_port = sp->s_port;
-      
-      biff_fd = socket (PF_INET, SOCK_DGRAM, 0);
-      if (biff_fd < 0)
-	{
-	  biff_fd = -2;
-	  return;
-	}
-    }
-
-  if (biff_fd)
-    {
-      mu_observer_create (&observer, mbox);
-      mu_observer_set_action (observer, notify_action, mbox);
-      mu_mailbox_get_observable (mbox, &observable);
-      mu_observable_attach (observable, MU_EVT_MESSAGE_APPEND, observer);
-    }
-}  
-
 int
-deliver_to_user (mu_mailbox_t mbox, mu_message_t msg,
-		 struct mu_auth_data *auth,
-		 char **errp)
+deliver_to_mailbox (mu_mailbox_t mbox, mu_message_t msg,
+		    struct mu_auth_data *auth,
+		    char **errp)
 {
   int status;
   char *path;
@@ -161,8 +150,6 @@ deliver_to_user (mu_mailbox_t mbox, mu_message_t msg,
       return EX_TEMPFAIL;
     }
 
-  attach_notify (mbox);
-  
   /* FIXME: This is superfluous, as mu_mailbox_append_message takes care
      of locking anyway. But I leave it here for the time being. */
   mu_mailbox_get_locker (mbox, &lock);
@@ -203,8 +190,8 @@ deliver_to_user (mu_mailbox_t mbox, mu_message_t msg,
 	  maidag_error (_("%s: mailbox quota exceeded for this recipient"),
 			auth->name);
 	  if (errp)
-	    asprintf (errp, "%s: mailbox quota exceeded for this recipient",
-		      auth->name);
+	    mu_asprintf (errp, "%s: mailbox quota exceeded for this recipient",
+		         auth->name);
 	  exit_code = EX_QUOTA();
 	  failed++;
 	  break;
@@ -226,10 +213,10 @@ deliver_to_user (mu_mailbox_t mbox, mu_message_t msg,
 			      "this recipient"),
 			    auth->name);
 	      if (errp)
-		asprintf (errp,
-			  "%s: message would exceed maximum mailbox size "
-			  "for this recipient",
-			  auth->name);
+		mu_asprintf (errp,
+			     "%s: message would exceed maximum mailbox size "
+			     "for this recipient",
+			     auth->name);
 	      exit_code = EX_QUOTA();
 	      failed++;
 	    }
@@ -268,42 +255,40 @@ deliver_to_user (mu_mailbox_t mbox, mu_message_t msg,
 static int
 is_remote_url (mu_url_t url)
 {
-  const char *scheme;
-  int rc = mu_url_sget_scheme (url, &scheme);
-  return rc == 0 && strncmp (scheme, "remote+", 7) == 0;
+  int rc, res;
+
+  if (!url)
+    return 0;
+  
+  rc = mu_registrar_test_local_url (url, &res);
+  return rc == 0 && res == 0;
 }
 
 static int
-is_mailer_url (mu_url_t url)
-{
-  mu_record_t record = NULL;
-  int (*pfn) (mu_mailer_t) = NULL;
-  
-  return mu_registrar_lookup_url (url, MU_FOLDER_ATTRIBUTE_FILE,
-				  &record, NULL) == 0
-         && mu_record_get_mailer (record, &pfn) == 0
-         && pfn;
-}
-
-int
-deliver_url (mu_url_t url, mu_message_t msg, const char *name, char **errp)
+do_delivery (mu_url_t url, mu_message_t msg, const char *name, char **errp)
 {
   struct mu_auth_data *auth = NULL;
   mu_mailbox_t mbox;
   int status;
 
-  if (name)
+  mu_set_user_email_domain (default_domain);
+  
+  if (name && !is_remote_url (url))
     {
       auth = mu_get_auth_by_name (name);
       if (!auth)
 	{
 	  maidag_error (_("%s: no such user"), name);
 	  if (errp)
-	    asprintf (errp, "%s: no such user", name);
+	    mu_asprintf (errp, "%s: no such user", name);
 	  exit_code = EX_NOUSER;
 	  return EX_NOUSER;
 	}
 
+      status = mu_set_user_email (name);
+      if (status)
+	mu_error (_("%s: invalid email: %s"), name, mu_strerror (status));
+      
       if (current_uid)
 	auth->change_uid = 0;
 
@@ -335,6 +320,8 @@ deliver_url (mu_url_t url, mu_message_t msg, const char *name, char **errp)
 	    return exit_code = EX_TEMPFAIL;
 	  }
     }
+  else
+    mu_set_user_email (NULL);
   
   if (!url)
     {
@@ -345,13 +332,6 @@ deliver_url (mu_url_t url, mu_message_t msg, const char *name, char **errp)
 			auth->mailbox, mu_strerror (status));
 	  return exit_code = EX_UNAVAILABLE;
 	}
-      status = mu_url_parse (url);
-      if (status)
-	{
-	  maidag_error (_("error parsing URL %s: %s"),
-			auth->mailbox, mu_strerror (status));
-	  return exit_code = EX_UNAVAILABLE;
-	}
     }      
 
   status = mu_mailbox_create_from_url (&mbox, url);
@@ -366,14 +346,18 @@ deliver_url (mu_url_t url, mu_message_t msg, const char *name, char **errp)
       return EX_TEMPFAIL;
     }
 
-  biff_user_name = name;
+  status = mu_mailbox_set_notify (mbox, name);
+  if (status)
+    mu_error (_("failed to set notification on %s: %s"),
+	      mu_url_to_string (url),
+	      mu_strerror (status));
   
   /* Actually open the mailbox. Switch to the user's euid to make
      sure the maildrop file will have right privileges, in case it
      will be created */
   if (switch_user_id (auth, 1))
     return EX_TEMPFAIL;
-  status = deliver_to_user (mbox, msg, auth, errp);
+  status = deliver_to_mailbox (mbox, msg, auth, errp);
   if (switch_user_id (auth, 0))
     return EX_TEMPFAIL;
 
@@ -384,56 +368,34 @@ deliver_url (mu_url_t url, mu_message_t msg, const char *name, char **errp)
 }
 
 int
-deliver (mu_message_t msg, char *dest_id, char **errp)
+deliver_to_url (mu_message_t msg, char *dest_id, char **errp)
 {
   int status;
   const char *name;
   mu_url_t url = NULL;
   
-  if (url_option)
+  status = mu_url_create (&url, dest_id);
+  if (status)
     {
-      status = mu_url_create (&url, dest_id);
-      if (status)
-	{
-	  maidag_error (_("%s: cannot create url: %s"), dest_id,
-			mu_strerror (status));
-	  return EX_NOUSER;
-	}
-      status = mu_url_parse (url);
-      if (status)
-	{
-	  maidag_error (_("%s: cannot parse url: %s"), dest_id,
-			mu_strerror (status));
-	  mu_url_destroy (&url);
-	  return EX_NOUSER;
-	}
-      status = mu_url_sget_user (url, &name);
-      if (status == MU_ERR_NOENT)
-	{
-	  if (!is_mailer_url (url) && !is_remote_url (url))
-	    {
-	      maidag_error (_("no user name"));
-	      if (errp)
-		asprintf (errp, "no such user");
-	      exit_code = EX_NOUSER;
-	      return EX_NOUSER;
-	    }
-	  else
-	    name = NULL;
-	}
-      else if (status)
-	{
-	  maidag_error (_("%s: cannot get user name from url: %s"),
-			dest_id, mu_strerror (status));
-	  mu_url_destroy (&url);
-	  return EX_NOUSER;
-	}
+      maidag_error (_("%s: cannot create url: %s"), dest_id,
+		    mu_strerror (status));
+      return EX_NOUSER;
     }
-  else
+  status = mu_url_sget_user (url, &name);
+  if (status == MU_ERR_NOENT)
+    name = NULL;
+  else if (status)
     {
-      name = dest_id;
-      dest_id = NULL;
+      maidag_error (_("%s: cannot get user name from url: %s"),
+		    dest_id, mu_strerror (status));
+      mu_url_destroy (&url);
+      return EX_NOUSER;
     }
-  return deliver_url (url, msg, name, errp);
+  return do_delivery (url, msg, name, errp);
 }
   
+int
+deliver_to_user (mu_message_t msg, char *dest_id, char **errp)
+{
+  return do_delivery (NULL, msg, dest_id, errp);
+}
