@@ -1,6 +1,6 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2000, 2001, 2002, 2005, 2007, 2009, 2010, 2011
-   Free Software Foundation, Inc.
+   Copyright (C) 1999-2002, 2005, 2007, 2009-2012, 2014-2016 Free
+   Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -50,15 +50,15 @@ pop3d_apopuser (const char *user)
     mu_dbm_file_t db;
     struct mu_dbm_datum key, data;
 
-    rc = mu_dbm_create (apop_database_name, &db);
+    rc = mu_dbm_create (apop_database_name, &db, apop_database_safety);
     if (rc)
       {
 	mu_diag_output (MU_DIAG_ERROR, _("unable to create APOP db"));
 	return NULL;
       }
 
-    if (apop_database_safety_set)
-      mu_dbm_safety_set_flags (db, apop_database_safety);
+    if (apop_database_owner_set)
+      mu_dbm_safety_set_owner (db, apop_database_owner);
     
     rc = mu_dbm_safety_check (db);
     if (rc)
@@ -114,12 +114,12 @@ pop3d_apopuser (const char *user)
 #else /* !ENABLE_DBM */
   {
     char *buf = NULL;
-    size_t size = 0;
+    size_t size = 0, n;
     size_t ulen;
-    FILE *apop_file;
+    mu_stream_t apop_stream;
 
     rc = mu_file_safety_check (apop_database_name, apop_database_safety,
-			       apop_database_uid, NULL);
+			       apop_database_owner, NULL);
     if (rc)
       {
 	mu_diag_output (MU_DIAG_ERROR,
@@ -127,17 +127,19 @@ pop3d_apopuser (const char *user)
 			apop_database_name, mu_strerror (rc));
 	return NULL;
       }
-    apop_file = fopen (apop_database_name, "r");
-    if (apop_file == NULL)
+
+    rc = mu_file_stream_create (&apop_stream, apop_database_name,
+				MU_STREAM_READ);
+    if (rc)
       {
 	mu_diag_output (MU_DIAG_INFO,
 			_("unable to open APOP password file %s: %s"),
-			apop_database_name, mu_strerror (errno));
+			apop_database_name, mu_strerror (rc));
 	return NULL;
       }
-
+    
     ulen = strlen (user);
-    while (getline (&buf, &size, apop_file) > 0)
+    while (mu_stream_getline (apop_stream, &buf, &size, &n) == 0 && n > 0)
       {
 	char *p, *start = mu_str_stripws (buf);
 
@@ -150,19 +152,19 @@ pop3d_apopuser (const char *user)
 	  {
 	    p = mu_str_skip_class (p + 1, MU_CTYPE_SPACE);
 	    if (*p)
-	      password = strdup (p);
+	      password = mu_strdup (p);
 	    break;
 	  }
       }
-
-    fclose (apop_file);
+    free (buf);
+    mu_stream_unref (apop_stream);
     return password;
   }
 #endif
 }
 
 int
-pop3d_apop (char *arg)
+pop3d_apop (char *arg, struct pop3d_session *sess)
 {
   char *p, *password, *user_digest, *user;
   struct mu_md5_ctx md5context;
